@@ -1,12 +1,12 @@
 # flake8: noqa
-__version__ = "0.0.17"
+__version__ = "0.0.18"
 
 from .base import ClosableResponse, EmptyResponse, Header, Method, Request, Response
 from .client import Client, setup
 from .context import get_context, set_context
 from .deadline import Deadline
 from .delays_provider import constant_delays, linear_delays
-from .metrics_collector import MetricsCollector
+from .metrics import ClientMetricsCollector
 from .priority import Priority
 from .request_sender import RequestSender
 from .requests import delete, get, post, post_json, put, put_json
@@ -23,6 +23,6 @@ except ImportError:
 try:
     import prometheus_client
 
-    from .prometheus import PrometheusMetricsCollector
+    from .prometheus import PrometheusClientMetricsCollector
 except ImportError:
     pass
