@@ -3,9 +3,9 @@ from typing import Optional
 import prometheus_client
 
 from .base import Request, Response
-from .metrics_collector import MetricsCollector
+from .metrics import ClientMetricsCollector
 
-requests_counter = prometheus_client.Counter(
+requests_status_counter = prometheus_client.Counter(
     name="aio_request_status",
     documentation="Response status",
     labelnames=("aio_request_service_name", "aio_request_method", "aio_request_path", "aio_request_status"),
@@ -18,7 +18,7 @@ requests_latency_histogram = prometheus_client.Histogram(
 )
 
 
-class PrometheusMetricsCollector(MetricsCollector):
+class PrometheusClientMetricsCollector(ClientMetricsCollector):
     __slots__ = ()
 
     def collect(self, request: Request, response: Optional[Response], elapsed_seconds: float) -> None:
@@ -28,5 +28,5 @@ class PrometheusMetricsCollector(MetricsCollector):
             request.url.path,
             response.status if response is not None else 499,
         )
-        requests_counter.labels(*labels).inc()
+        requests_status_counter.labels(*labels).inc()
         requests_latency_histogram.labels(*labels).observe(elapsed_seconds)
